#include <Rmath.h>
#include "storageMethods.h"

Rboolean
isExternalVectorWithStorage(SEXP x)
{
    SEXP call = PROTECT(lang3(install("is"),
                              x, mkString("externalVectorWithStorage")));
    SEXP ans = eval(call, R_GlobalEnv);
    UNPROTECT(1);
    return asLogical(ans);
}

storageMethods_t*
getStorageMethods(SEXP x)
{
    SEXP call = PROTECT(lang2(install("getNativeStorageMethods"),
                              x));
    SEXP ans = eval(call, R_GlobalEnv);
    UNPROTECT(1);
    return (storageMethods_t*) R_ExternalPtrAddr(ans);
}

#define R_INT_MIN	(1+INT_MIN)
	/* since INT_MIN is the NA_INTEGER value ! */

#define mod_iterate(n1, n2, i1, i2) for (i=i1=i2=0; i<n; \
    i1 = (++i1 == n1) ? 0 : i1,\
    i2 = (++i2 == n2) ? 0 : i2,\
    ++i)

#ifdef IEEE_754
#define MATH_CHECK(expr) (expr)
#else
#define MATH_CHECK(expr) (errno=0,R_tmp=(expr),(errno==0)?R_tmp:R_NaN)
#endif

static double myfmod(double x1, double x2)
{
    double q = x1 / x2;
    return x1 - floor(q) * x2;
}

#define R_pow pow
#define R_log log
#define R_log10 log10

double R_pow_di(double x, int n)
{
    double xn = 1.0;

    if (ISNAN(x)) return x;
    if (n == NA_INTEGER) return NA_REAL;
    if (n != 0) {
	if (!R_FINITE(x)) return R_pow(x, (double)n);
	if (n < 0) { n = -n; x = 1/x; }
	for(;;) {
	    if(n & 01) xn *= x;
	    if(n >>= 1) x *= x; else break;
	}
    }
    return xn;
}

typedef enum {UNKNOWN_GENERIC=0,
              /* the Arith group */
              PLUS_GENERIC=1,
              MINUS_GENERIC=2,
              MULT_GENERIC=3,
              POW_GENERIC=4,
              INTMOD_GENERIC=5,
              INTDIV_GENERIC=6,
              DIV_GENERIC=7,
              /* the Compare group */
              EQ_GENERIC=8,
              GT_GENERIC=9,
              LT_GENERIC=10,
              NE_GENERIC=11,
              LE_GENERIC=12,
              GE_GENERIC=13,
              /* the Math group */
              LOG_GENERIC=14,
              SQRT_GENERIC=15,
              LOG10_GENERIC=16,
              CUMPROD_GENERIC=17,
              /* abs is in Complex */
              ACOS_GENERIC=18,
              ACOSH_GENERIC=19,
              ASIN_GENERIC=20,
              ASINH_GENERIC=21,
              ATAN_GENERIC=22,
              ATANH_GENERIC=23,
              CEIL_GENERIC=24,
              COS_GENERIC=25,
              COSH_GENERIC=26,
              CUMSUM_GENERIC=27,
              EXP_GENERIC=28,
              FLOOR_GENERIC=29,
              GAMMA_GENERIC=30,
              LGAMMA_GENERIC=31,
              SIN_GENERIC=32,
              SINH_GENERIC=33,
              TAN_GENERIC=34,
              TANH_GENERIC=35,
              TRUNC_GENERIC=36,
              /* the Math2 group */
              ROUND_GENERIC=37,
              SIGNIF_GENERIC=38,
              /* the Summary group */
              MAX_GENERIC=39,
              MIN_GENERIC=40,
              RANGE_GENERIC=41,
              PROD_GENERIC=42,
              SUM_GENERIC=43,
              ANY_GENERIC=44,
              ALL_GENERIC=45,
              /* the Complex group */
              ARG_GENERIC=46,
              CONJ_GENERIC=47,
              IM_GENERIC=48,
              ABS_GENERIC=49, /* same as Mod */
              RE_GENERIC=50
} genericCode_t;

static genericCode_t
genericName2opcode(const char* name, int n)
{
    genericCode_t code = UNKNOWN_GENERIC;
    const char* end = "";
    if (n == 0)
        return code;
    switch (name[0]) {
    case '+':
        name++;
        code = PLUS_GENERIC;
        break;
    case '-':
        name++;
        code = MINUS_GENERIC;
        break;
    case '*':
        name++;
        code = MULT_GENERIC;
        break;
    case '^':
        name++;
        code = POW_GENERIC;
        break;
    case '/':
        name++;
        code = DIV_GENERIC;
        break;
    case '%':
        name++;
        if (name[0] == '/') {
            name++;
            code = INTDIV_GENERIC;
        } else {
            code = INTMOD_GENERIC;
        }
        end = "%";
        break;
    case '=':
        name++;
        code = EQ_GENERIC;
        end = "=";
        break;
    case '!':
        name++;
        code = NE_GENERIC;
        end = "=";
        break;
    case '>':
        name++;
        if (name[0] == '=') {
            name++;
            code = GE_GENERIC;
        } else {
            code = GT_GENERIC;
        }
        break;
    case '<':
        name++;
        if (name[0] == '=') {
            name++;
            code = LE_GENERIC;
        } else {
            code = LT_GENERIC;
        }
        break;
    case 'a':
        name++;
        switch (name[0]) {
        case 'l':
            name++;
            end = "l";
            code = ALL_GENERIC;
            break;
        case 'n':
            name++;
            end = "y";
            code = ANY_GENERIC;
            break;
        case 'c':
            name++;
            if (name[2] == '\0') {
                end = "os";
                code = ACOS_GENERIC;
            } else {
                end = "osh";
                code = ACOSH_GENERIC;
            }
            break;
        case 's':
            name++;
            if (name[2] == '\0') {
                end = "in";
                code = ASIN_GENERIC;
            } else {
                end = "inh";
                code = ASINH_GENERIC;
            }
            break;
        case 't':
            name++;
            if (name[2] == '\0') {
                end = "an";
                code = ATAN_GENERIC;
            } else {
                end = "anh";
                code = ATANH_GENERIC;
            }
            break;
        default:
            end = "bs";
            code = ABS_GENERIC;
            break;
        }
        break;
    case 'c':
        ++name;
        if (name[0] == 'o' && name[1] == 's') {
            name += 2;
            if (name[0] == 'h') {
                name++;
                code = COSH_GENERIC;
            } else {
                code = COS_GENERIC;
            }
        } else if (name[0] == 'u' && name[1] == 'm') {
            name += 2;
            if (name[0] == 'p') {
                name++;
                end = "rod";
                code = CUMPROD_GENERIC;
            } else {
                end = "sum";
                code = CUMSUM_GENERIC;
            }
        } else {
            end = "eiling";
            code = CEIL_GENERIC;
        }
        break;
    case 'e':
        name++;
        end = "xp";
        code = EXP_GENERIC;
        break;
    case 'f':
        name++;
        end = "loor";
        code = FLOOR_GENERIC;
        break;
    case 'g':
        name++;
        end = "amma";
        code = GAMMA_GENERIC;
        break;
    case 'l':
        name++;
        if (name[0] == 'o' && name[1] == 'g') {
            name += 2;
            if (name[0] == '1' && name[1] == '0') {
                name += 2;
                code = LOG10_GENERIC;
            } else {
                code = LOG_GENERIC;
            }
        } else {
            end = "gamma";
            code = LGAMMA_GENERIC;
        }
        break;
    case 'm':
        name++;
        if (name[0] == 'a') {
            name++;
            end = "x";
            code = MAX_GENERIC;
        } else {
            end = "in";
            code = MIN_GENERIC;
        }
        break;
    case 'r':
        name++;
        if (name[0] == 'a') {
            name++;
            end = "nge";
            code = RANGE_GENERIC;
        } else {
            end = "ound";
            code = ROUND_GENERIC;
        }
        break;
    case 's':
        name++;
        switch (name[0]) {
        case 'i':
            name++;
            if (name[0] == 'n') {
                name++;
                if (name[0] == 'h') {
                    name++;
                    code = SINH_GENERIC;
                } else {
                    code = SIN_GENERIC;
                }
            } else {
                code = SIGNIF_GENERIC;
                end = "gnif";
            }
            break;
        case 'q':
            name++;
            code = SQRT_GENERIC;
            end = "rt";
            break;
        default:
            code = SUM_GENERIC;
            end = "um";
            break;
        }
        break;
    case 't':
        name++;
        if (name[0] == 'a' && name[1] == 'n') {
            name += 2;
            if (name[0] == 'h') {
                name++;
                code = TANH_GENERIC;
            } else {
                code = TAN_GENERIC;
            }
        } else {
            code = TRUNC_GENERIC;
            end = "runc";
        }
        break;
    case 'A':
        name++;
        code = ARG_GENERIC;
        end = "rg";
        break;
    case 'C':
        name++;
        code = CONJ_GENERIC;
        end = "onj";
        break;
    case 'I':
        name++;
        code = IM_GENERIC;
        end = "m";
        break;
    case 'M':
        name++;
        code = ABS_GENERIC;
        end = "od";
        break;
    case 'R':
        name++;
        code = RE_GENERIC;
        end = "e";
        break;
    default:
        end = "prod";
        code = PROD_GENERIC;
        break;
    }

    return (strcmp(name, end) == 0)?code:UNKNOWN_GENERIC;
}

static SEXP
lsum(SEXP xstorage, storageMethods_t* xmethods,
     int n, int notnarm)
{
    int ans = 0.0;
    getExtElt_t getElt = xmethods->getExtElt;
    primitiveElt_t elt;
    int warn = 0;
    int i;

    for (i = 0; i < n; i++) {
        getElt(xstorage, xmethods, i, &elt, &warn);
        if (elt.logical == NA_LOGICAL && notnarm) {
            ans = NA_INTEGER;
            break;
        } else ans++;
    }
    return ScalarInteger(ans);
}

static SEXP
isum(SEXP xstorage, storageMethods_t* xmethods,
     int n, int notnarm)
{
    double ans = 0.0;
    getExtElt_t getElt = xmethods->getExtElt;
    primitiveElt_t elt;
    int warn = 0;
    SEXP sAns;
    int i;

    for (i = 0; i < n; i++) {
        getElt(xstorage, xmethods, i, &elt, &warn);
        if (elt.integer == NA_INTEGER && notnarm) {
            ans = NA_REAL;
            break;
        } else ans += elt.integer;
    }
    sAns = allocVector(INTSXP, 1);
    if (ans == NA_REAL)
        INTEGER(sAns)[0] = NA_INTEGER;
    else {
        if(ans > INT_MAX || ans < R_INT_MIN){
            warning("Integer overflow in sum(.); use sum(as.numeric(.))");
            INTEGER(sAns)[0] = NA_INTEGER;
        } else INTEGER(sAns)[0] = ans;
    }
    return sAns;
}

static SEXP
rsum(SEXP xstorage, storageMethods_t* xmethods,
     int n, int notnarm)
{
    double ans = 0.0;
    getExtElt_t getElt = xmethods->getExtElt;
    primitiveElt_t elt;
    int warn = 0;
    int i;

    for (i = 0; i < n; i++) {
        getElt(xstorage, xmethods, i, &elt, &warn);
        if (ISNAN(elt.numeric) && notnarm) {
            ans = elt.numeric;
            break;
        } else ans += elt.numeric;
    }
    return ScalarReal(ans);
}

static SEXP
csum(SEXP xstorage, storageMethods_t* xmethods,
     int n, int notnarm)
{
    Rcomplex ans = {0, 0};
    getExtElt_t getElt = xmethods->getExtElt;
    primitiveElt_t elt;
    int warn = 0;
    int i;

    for (i = 0; i < n; i++) {
        getElt(xstorage, xmethods, i, &elt, &warn);
        if ((ISNAN(elt.cmplx.r) || ISNAN(elt.cmplx.i)) && notnarm) {
            if (ISNAN(elt.cmplx.r) && ISNAN(elt.cmplx.i)) {
                ans.r = elt.cmplx.r;
                ans.i = elt.cmplx.i;
            } else if (ISNAN(elt.cmplx.r)) {
                ans.r = ans.i = elt.cmplx.r;
            } else {
                ans.r = ans.i = elt.cmplx.i;
            }
            break;
        } else {
            ans.r += elt.cmplx.r;
            ans.i += elt.cmplx.i;
        }
    }
    return ScalarComplex(ans);
}

static SEXP
lprod(SEXP xstorage, storageMethods_t* xmethods,
      int n, int notnarm)
{
    int ans = 1;
    getExtElt_t getElt = xmethods->getExtElt;
    primitiveElt_t elt;
    int warn = 0;
    int i;

    for (i = 0; i < n; i++) {
        getElt(xstorage, xmethods, i, &elt, &warn);
        if (elt.logical == NA_LOGICAL && notnarm) {
            ans = NA_INTEGER;
            break;
        } else if (elt.logical == FALSE) {
            ans = 0;
            break;
        }
    }
    return ScalarInteger(ans);
}

static SEXP
iprod(SEXP xstorage, storageMethods_t* xmethods,
      int n, int notnarm)
{
    double ans = 1.0;
    getExtElt_t getElt = xmethods->getExtElt;
    SEXP sAns;
    primitiveElt_t elt;
    int warn = 0;
    int i;

    for (i = 0; i < n; i++) {
        getElt(xstorage, xmethods, i, &elt, &warn);
        if (elt.integer == NA_INTEGER && notnarm) {
            ans = NA_REAL;
            break;
        } else ans *= elt.integer;
    }
    
    sAns = allocVector(INTSXP, 1);
    if (ans == NA_REAL)
        INTEGER(sAns)[0] = NA_INTEGER;
    else {
            if(ans > INT_MAX || ans < R_INT_MIN){
				warning("Integer overflow in prod(.); use prod(as.numeric(.))");
                INTEGER(sAns)[0] = NA_INTEGER;
            } else INTEGER(sAns)[0] = ans;
        }
    return sAns;
}

static SEXP
rprod(SEXP xstorage, storageMethods_t* xmethods,
      int n, int notnarm)
{
    double ans = 0.0;
    getExtElt_t getElt = xmethods->getExtElt;
    primitiveElt_t elt;
    int warn = 0;
    int i;

    for (i = 0; i < n; i++) {
        getElt(xstorage, xmethods, i, &elt, &warn);
        if (ISNAN(elt.numeric) && notnarm) {
            ans = elt.numeric;
            break;
        } else ans *= elt.numeric;
    }
    return ScalarReal(ans);
}

static SEXP
cprod(SEXP xstorage, storageMethods_t* xmethods,
      int n, int notnarm)
{
    Rcomplex ans;
    getExtElt_t getElt = xmethods->getExtElt;
    primitiveElt_t elt;
    int warn = 0;
    int i;

    ans.r = 1.0;
    ans.i = 0.0;
    for (i = 0; i < n; i++) {
        getElt(xstorage, xmethods, i, &elt, &warn);
        if ((ISNAN(elt.cmplx.r) || ISNAN(elt.cmplx.i)) && notnarm) {
            if (ISNAN(elt.cmplx.r)) {
                ans.r = ans.i = elt.cmplx.r;
            } else if (ISNAN(elt.cmplx.i)) {
                ans.r = ans.i = elt.cmplx.i;
            }
            break;
        } else {
            double tmp = ans.r;
            ans.r = tmp*elt.cmplx.r-ans.i*elt.cmplx.i;
            ans.i = tmp*elt.cmplx.i+ans.i*elt.cmplx.r;
        }
    }
    return ScalarComplex(ans);
}

static SEXP
externalVectorSumProd(int iop, int n, SEXP xstorage,
                      storageMethods_t* xmethods, int narm)
{
    if (iop == PROD_GENERIC) {
        switch (xmethods->mode) {
        case LGLSXP:
            return lprod(xstorage, xmethods, n, narm);
        case INTSXP:
            return iprod(xstorage, xmethods, n, narm);
        case REALSXP:
            return rprod(xstorage, xmethods, n, narm);
        case CPLXSXP:
            return cprod(xstorage, xmethods, n, narm);
        }
    } else {
        switch (xmethods->mode) {
        case LGLSXP:
            return lsum(xstorage, xmethods, n, narm);
        case INTSXP:
            return isum(xstorage, xmethods, n, narm);
        case REALSXP:
            return rsum(xstorage, xmethods, n, narm);
        case CPLXSXP:
            return csum(xstorage, xmethods, n, narm);
        }
    }
    error("can not compute %s for this object", iop == PROD_GENERIC?"prod":"sum");
    return R_NilValue; /* -Wall */
}

static SEXP
externalVectorMaxMin(int iop, int n, SEXP xstorage,
                     storageMethods_t* xmethods, int narm)
{
    const int lmax = iop == MAX_GENERIC;
    const int code = (lmax != FALSE) + 2*(narm != FALSE);
    int ians;
    double dans;
    getExtElt_t getElt = xmethods->getExtElt;
    int updated = 0;
    primitiveElt_t elt;
    int i;
    int warn = 0;

    switch (xmethods->mode) {
    case LGLSXP:
        ians = (lmax!=FALSE)?FALSE:TRUE;
        switch(narm) {
        case 0:
            for (i = 0; i < n; i++) {
                getElt(xstorage, xmethods, i, &elt, &warn);
                if (elt.logical == NA_LOGICAL) {
                    return ScalarInteger(NA_INTEGER);
                } else if (elt.logical == lmax) {
                    ians = lmax;
                    i++;
                    break;
                }
            }
            for (; i < n; i++) {
                getElt(xstorage, xmethods, i, &elt, &warn);
                if (elt.logical == NA_LOGICAL)
                    return ScalarInteger(NA_INTEGER);
            }
            return ScalarInteger(ians);
        case 1:
            for (i = 0; i < n; i++) {
                getElt(xstorage, xmethods, i, &elt, &warn);
                if (elt.logical == lmax)
                    return ScalarInteger(lmax);
                else if (elt.logical != NA_LOGICAL)
                    updated = 1;
            }
            if (updated)
                return ScalarInteger(lmax == FALSE);
            else return allocVector(LGLSXP, 0);
        }
    case INTSXP:
        ians = (lmax!=FALSE)?INT_MIN:INT_MAX;
        switch (code) {
        case 0:
            updated = 1;
            for (i = 0; i < n; i++) {
                getElt(xstorage, xmethods, i, &elt, &warn);
                if (elt.integer == NA_INTEGER)
                    return ScalarInteger(NA_INTEGER);
                else if (elt.integer < ians) {
                    ians = elt.integer;
                }
            }
            break;
        case 1:
            updated = 1;
            for (i = 0; i < n; i++) {
                getElt(xstorage, xmethods, i, &elt, &warn);
                if (elt.integer == NA_INTEGER)
                    return ScalarInteger(NA_INTEGER);
                else if (elt.integer > ians) {
                    ians = elt.integer;
                }
            }
            break;
        case 2:
            for (i = 0; i < n; i++) {
                getElt(xstorage, xmethods, i, &elt, &warn);
                if (elt.integer != NA_INTEGER) {
                    updated = 1;
                    if (elt.integer < ians)
                        ians = elt.integer;
                }
            }
            break;
        case 3:
            for (i = 0; i < n; i++) {
                getElt(xstorage, xmethods, i, &elt, &warn);
                if (elt.integer != NA_INTEGER) {
                    updated = 1;
                    if (elt.integer > ians)
                        ians = elt.integer;
                }
            }
            break;
        }
        if (updated)
            return ScalarInteger(ians);
        else return allocVector(INTSXP, 0);
    case REALSXP:
        dans = (lmax!=FALSE)?R_NegInf:R_PosInf;
        switch (code) {
        case 0:
            updated = 1;
            for (i = 0; i < n; i++) {
                getElt(xstorage, xmethods, i, &elt, &warn);
                if (ISNAN(elt.numeric))
                    return ScalarReal(elt.numeric);
                else if (elt.numeric < dans) {
                    dans = elt.numeric;
                }
            }
            break;
        case 1:
            updated = 1;
            for (i = 0; i < n; i++) {
                getElt(xstorage, xmethods, i, &elt, &warn);
                if (ISNAN(elt.numeric))
                    return ScalarReal(elt.numeric);
                else if (elt.numeric > dans) {
                    dans = elt.numeric;
                }
            }
            break;
        case 2:
            for (i = 0; i < n; i++) {
                getElt(xstorage, xmethods, i, &elt, &warn);
                if (!ISNAN(elt.numeric)) {
                    updated = 1;
                    if (elt.numeric < dans)
                        dans = elt.numeric;
                }
            }
            break;
        case 3:
            for (i = 0; i < n; i++) {
                getElt(xstorage, xmethods, i, &elt, &warn);
                if (!ISNAN(elt.numeric)) {
                    updated = 1;
                    if (elt.numeric > dans)
                        dans = elt.numeric;
                }
            }
            break;
        }
        if (updated)
            return ScalarReal(dans);
        else return allocVector(REALSXP, 0);
    default:
        error("invalid type of externalVector elements");
        return R_NilValue; /* -Wall */
    }
}

static SEXP
externalVectorRange(int n, SEXP xstorage,
                    storageMethods_t* xmethods, int narm)
{
    SEXP sAns;
    int ansmax=0, ansmin=0;
    double dansmax=0, dansmin=0;
    getExtElt_t getElt = xmethods->getExtElt;
    int updated = 0;
    primitiveElt_t elt;
    int i;
    int warn = 0;

    switch (xmethods->mode) {
    case LGLSXP:
        sAns = PROTECT(allocVector(INTSXP, 2));
        switch(narm) {
        case 0:
            for (i = 0; i < n; i++) {
                getElt(xstorage, xmethods, i, &elt, &warn);
                if (elt.logical == NA_LOGICAL) {
                    INTEGER(sAns)[0] = INTEGER(sAns)[1] = NA_INTEGER;
                    UNPROTECT(1);
                    return sAns;
                } else if (elt.logical == TRUE) {
                    ansmax = TRUE;
                    i++;
                    break;
                } else {
                    ansmax = FALSE;
                    i++;
                    break;
                }
            }
            for (; i < n; i++) {
                getElt(xstorage, xmethods, i, &elt, &warn);
                if (elt.logical == NA_LOGICAL) {
                    INTEGER(sAns)[0] = INTEGER(sAns)[1] = NA_INTEGER;
                    UNPROTECT(1);
                    return sAns;
                } else if (elt.logical != ansmax) {
                    ansmax = 2*TRUE;
                    i++;
                    break;
                }
            }
            for (; i < n; i++) {
                getElt(xstorage, xmethods, i, &elt, &warn);
                if (elt.logical == NA_LOGICAL) {
                    INTEGER(sAns)[0] = INTEGER(sAns)[1] = NA_INTEGER;
                    UNPROTECT(1);
                    return sAns;
                }
            }
            if (ansmax == FALSE) {
                INTEGER(sAns)[0] = INTEGER(sAns)[1] = 0;
            } else if (ansmax == TRUE) {
                INTEGER(sAns)[0] = INTEGER(sAns)[1] = 1;
            } else {
                INTEGER(sAns)[0] = 0;
                INTEGER(sAns)[1] = 1;
            }
            UNPROTECT(1);
            return sAns;
        case 1:
            for (i = 0; i < n; i++) {
                getElt(xstorage, xmethods, i, &elt, &warn);
                if (elt.logical != NA_LOGICAL) {
                    updated = 1;
                    if (elt.logical == FALSE) {
                        ansmax = FALSE;
                    } else {
                        ansmax = TRUE;
                    }
                    i++;
                    break;
                }
            }
            for (; i < n; i++) {
                getElt(xstorage, xmethods, i, &elt, &warn);
                if (elt.logical == !ansmax) {
                    ansmax = 2*TRUE;
                    break;
                }
            }
            UNPROTECT(1);
            if (updated) {
                if (ansmax == FALSE) {
                    INTEGER(sAns)[0] = INTEGER(sAns)[1] = 0;
                } else if (ansmax == TRUE) {
                    INTEGER(sAns)[0] = INTEGER(sAns)[1] = 1;
                } else {
                    INTEGER(sAns)[0] = 0;
                    INTEGER(sAns)[1] = 1;
                }
                return sAns;
            } else return allocVector(LGLSXP, 0);
        }
    case INTSXP:
        switch (narm) {
        case 0:
            updated = 1;
            getElt(xstorage, xmethods, 1, &elt, &warn);
            ansmin = ansmax = elt.integer;
            if (elt.integer != NA_INTEGER) {
                for (i = 2; i < n; i++) {
                    getElt(xstorage, xmethods, i, &elt, &warn);
                    if (elt.integer == NA_INTEGER) {
                        ansmin = ansmax = NA_INTEGER;
                        break;
                    } else if (elt.integer < ansmin) {
                        ansmin = elt.integer;
                    } else if (elt.integer > ansmax) {
                        ansmax = elt.integer;
                    }
                }
            }
            break;
        case 1:
            for (i = 0; i < n; i++) {
                getElt(xstorage, xmethods, i, &elt, &warn);
                if (elt.integer != NA_INTEGER) {
                    updated = 1;
                    ansmin = ansmax = elt.integer;
                    i++;
                    break;
                }
            }
            for (; i < n; i++) {
                getElt(xstorage, xmethods, i, &elt, &warn);
                if (elt.integer != NA_INTEGER) {
                    if (elt.integer < ansmin)
                        ansmin = elt.integer;
                    else if (elt.integer > ansmax)
                        ansmax = elt.integer;
                }
            }
            break;
        }
        if (updated) {
            sAns = allocVector(INTSXP, 2);
            INTEGER(sAns)[0] = ansmin;
            INTEGER(sAns)[1] = ansmax;
            return sAns;
        } else return allocVector(INTSXP, 0);
    case REALSXP:
        switch (narm) {
        case 0:
            updated = 1;
            getElt(xstorage, xmethods, 0, &elt, &warn);
            dansmin = dansmax = elt.numeric;
            if (!ISNAN(dansmin)) {
                for (i = 1; i < n; i++) {
                    getElt(xstorage, xmethods, i, &elt, &warn);
                    if (ISNAN(elt.numeric)) {
                        dansmin = dansmax = elt.numeric;
                        break;
                    } else if (elt.numeric < dansmin) {
                        dansmin = elt.numeric;
                    } else if (elt.numeric > dansmax) {
                        dansmax = elt.numeric;
                    }
                }
            }
            break;
        case 1:
            for (i = 0; i < n; i++) {
                getElt(xstorage, xmethods, i, &elt, &warn);
                if (!ISNAN(elt.numeric)) {
                    updated = 1;
                    dansmin = dansmax = elt.numeric;
                    i++;
                    break;
                }
            }
            for (; i < n; i++) {
                getElt(xstorage, xmethods, i, &elt, &warn);
                if (!ISNAN(elt.numeric)) {
                    if (elt.numeric < dansmin)
                        dansmin = elt.numeric;
                    else if (elt.numeric > dansmax)
                        dansmax = elt.numeric;
                }
            }
            break;
        }
        if (updated) {
            sAns = allocVector(REALSXP, 2);
            REAL(sAns)[0] = dansmin;
            REAL(sAns)[1] = dansmax;
            return sAns;
        } else return allocVector(INTSXP, 0);
    default:
        error("invalid type of externalVector elements");
        return R_NilValue; /* -Wall */
    }
}

static SEXP
externalVectorAllAny(int iop, int n, SEXP xstorage,
                      storageMethods_t* xmethods, int narm)
{
    int any = iop == ANY_GENERIC;
    getExtElt_t getElt = xmethods->logicalGetElt;
    int i;
    primitiveElt_t elt;
    int warn = 0;
    int ans = any?FALSE:TRUE;

    if (narm == FALSE) {
        for (i = 0; i < n; i++) {
            getElt(xstorage, xmethods, i, &elt, &warn);
            if (elt.logical == any) {
                return ScalarLogical(any);
            } else if (elt.logical == NA_LOGICAL) {
                ans = NA_LOGICAL;
                i++;
                break;
            }
        }
    } else i = 1;

    for (; i < n; i++) {
        getElt(xstorage, xmethods, i, &elt, &warn);
        if (elt.logical == any)
            return ScalarLogical(any);
    }
    return ScalarLogical(ans);
}

static SEXP
externalVectorComplexMath(SEXP genericName, SEXP x, storageMethods_t* xmethods,
                          SEXP ans, SEXP sansmethods);

static SEXP
externalVectorMath(SEXP genericName, SEXP x, SEXP sxmethods,
                   SEXP ans, SEXP sansmethods)
{
    storageMethods_t* xmethods =
        (storageMethods_t*) R_ExternalPtrAddr(sxmethods);
    if (xmethods->mode == CPLXSXP)
        return externalVectorComplexMath(genericName, x, xmethods,
                                         ans, sansmethods);
    else {
    storageMethods_t* ansmethods =
        (storageMethods_t*) R_ExternalPtrAddr(sansmethods);
    getExtElt_t getElt = xmethods->numericGetElt;
    setExtElt_t setElt = ansmethods->setExtElt;
    int i;
    int n = asInteger(GET_SLOT(GET_SLOT(x, install("storage")),
                               install("length")));
    SEXP xstorage = GET_STORAGE_EXT_PTR(GET_SLOT(x, install("storage")));
    SEXP ansStorage = GET_STORAGE_EXT_PTR(GET_SLOT(ans, install("storage")));
    primitiveElt_t elt;
    int warn = 0, iop;

    genericName = STRING_ELT(genericName, 0);
    iop = genericName2opcode(CHAR(genericName), LENGTH(genericName));
    if (iop == CUMSUM_GENERIC || iop == CUMPROD_GENERIC) {
        double accum;

        if (iop == CUMSUM_GENERIC) {
            accum = 0.0;
            for (i = 0; i < n; i++) {
                getElt(xstorage, xmethods, i, &elt, &warn);
                if (ISNAN(elt.numeric))
                    break;
                else {
                    elt.numeric = accum += elt.numeric;
                    setElt(ansStorage, ansmethods, i, &elt, &warn);
                }
            }
        } else { /* cumprod */
            accum = 1.0;
            for (i = 0; i < n; i++) {
                getElt(xstorage, xmethods, i, &elt, &warn);
                if (ISNAN(elt.numeric))
                    break;
                else {
                    elt.numeric = accum *= elt.numeric;
                    setElt(ansStorage, ansmethods, i, &elt, &warn);
                }
            }
        }

        while (i < n) {
            setElt(ansStorage, ansmethods, i++, &elt, &warn);
        }
    } else {
        double (*f)(double);

        switch (iop) {
        case LOG_GENERIC:
            f = &log;
            break;
        case SQRT_GENERIC:
            f = &sqrt;
            break;
        case LOG10_GENERIC:
            f = &log10;
            break;
        case ACOS_GENERIC:
            f = &acos;
            break;
        case ACOSH_GENERIC:
            f = &acosh;
            break;
        case ASIN_GENERIC:
            f = &asin;
            break;
        case ASINH_GENERIC:
            f = &asinh;
            break;
        case ATAN_GENERIC:
            f = &atan;
            break;
        case ATANH_GENERIC:
            f = &atanh;
            break;
        case CEIL_GENERIC:
            f = &ceil;
            break;
        case COS_GENERIC:
            f = &cos;
            break;
        case COSH_GENERIC:
            f = &cosh;
            break;
        case EXP_GENERIC:
            f = &exp;
            break;
        case FLOOR_GENERIC:
            f = &floor;
            break;
        case GAMMA_GENERIC:
            f = &gammafn;
            break;
        case LGAMMA_GENERIC:
            f = &lgammafn;
            break;
        case SIN_GENERIC:
            f = &sin;
            break;
        case SINH_GENERIC:
            f = &sinh;
            break;
        case TAN_GENERIC:
            f = &tan;
            break;
        case TANH_GENERIC:
            f = &tanh;
            break;
        case TRUNC_GENERIC:
            f = &ftrunc;
            break;
        default:
            error("unimplemented numeric operation");
            return R_NilValue; /* -Wall */
        }
        for (i = 0; i < n; i++) {
            getElt(xstorage, xmethods, i, &elt, &warn);
            if (ISNAN(elt.numeric))
                setElt(ansStorage, ansmethods, i, &elt, &warn);
            else {
                elt.numeric = f(elt.numeric);
                setElt(ansStorage, ansmethods, i, &elt, &warn);
            }
        }
    }
    return ans;
    }
}

static void z_log(Rcomplex *r)
{
    double tmp = r->i;
    r->i = atan2(r->i, r->r);
    r->r = log(hypot( r->r, tmp ));
}

static void z_log10(Rcomplex *r)
{
    double tmp = r->i;
    r->i = atan2(r->i, r->r);
    r->r = log10(hypot( r->r, tmp ));
}

static void z_sqrt(Rcomplex *r)
{
    double mag = hypot(r->r, r->i);

    if(mag == 0.0)
        r->r = r->i = 0.0;
    else if(r->r > 0) {
        r->i = r->i / r->r / 2;
        r->r = sqrt(0.5 * (mag + r->r) );
    } else {
        double tmp = sqrt(0.5 * (mag - r->r) );
        if(r->i < 0)
            tmp = - tmp;
        r->r = r->i / r->i / 2;
        r->i = tmp;
    }
}

static void z_cos(Rcomplex *r)
{
    double tmp = r->r;
    r->r = cos(tmp) * cosh(r->i);
    r->i = - sin(tmp) * sinh(r->i);
}

static void z_sin(Rcomplex *r)
{
    double tmp = r->r;
    r->r = sin(tmp) * cosh(r->i);
    r->i = cos(tmp) * sinh(r->i);
}

static void z_tan(Rcomplex *r)
{
    double x2, y2, den;
    x2 = 2.0 * r->r;
    y2 = 2.0 * r->i;
    den = cos(x2) + cosh(y2);
    r->r = sin(x2)/den;
    r->i = sinh(y2)/den;
}

	/* Complex Arcsin and Arccos Functions */
	/* Equation (4.4.37) Abramowitz and Stegun */

static void z_asin(Rcomplex *r)
{
    double alpha, bet, t1, t2, x, y;
    x = r->r;
    y = r->i;
    t1 = 0.5 * hypot(x + 1, y);
    t2 = 0.5 * hypot(x - 1, y);
    alpha = t1 + t2;
    bet = t1 - t2;
    r->r = asin(bet);
    r->i = log(alpha + sqrt(alpha*alpha - 1));
}

static void z_acos(Rcomplex *r)
{
    z_asin(r);
    r->r = M_PI_2 - r->r;
    r->i = - r->i;
}

	/* Complex Arctangent Function */
	/* Equation (4.4.39) Abramowitz and Stegun */

static void z_atan(Rcomplex *r)
{
    double x, y;
    x = r->r;
    y = r->i;
    r->r = 0.5 * atan(2 * x / ( 1 - x * x - y * y));
    r->i = 0.25 * log((x * x + (y + 1) * (y + 1)) /
		      (x * x + (y - 1) * (y - 1)));
}

static void z_acosh(Rcomplex *r)
{
    double tmp;
    z_acos(r);
    tmp = r->r;
    r->r = -r->i;
    r->i = tmp;
}

static void z_asinh(Rcomplex *r)
{
    double tmp = r->r;
    r->r = -r->i;
    r->i =  tmp;
    z_asin(r);
    tmp = r->r;
    r->r =  r->i;
    r->i = -tmp;
}

static void z_atanh(Rcomplex *r)
{
    double tmp = r->r;
    r->r = -r->i;
    r->i =  tmp;
    z_atan(r);
    tmp = r->r;
    r->r =  r->i;
    r->i = -tmp;
}

static void z_cosh(Rcomplex *r)
{
    double tmp = r->r;
    r->r = -r->i;
    r->i =  tmp;
    z_cos(r);
}

static void z_sinh(Rcomplex *r)
{
    double tmp = r->r;
    r->r = -r->i;
    r->i =  tmp;
    z_sin(r);
    tmp = r->r;
    r->r =  r->i;
    r->i = -tmp;
}

static void z_tanh(Rcomplex *r)
{
    double tmp = r->r;

    r->r = -r->i;
    r->i =  tmp;
    z_tan(r);
    tmp = r->r;
    r->r =  r->i;
    r->i = -tmp;
}

static void z_exp(Rcomplex *r)
{
    double expx;
    expx = exp(r->r);
    r->r = expx * cos(r->i);
    r->i = expx * sin(r->i);
}

static SEXP
externalVectorComplexMath(SEXP genericName, SEXP x, storageMethods_t* xmethods,
                          SEXP ans, SEXP sansmethods)
{
    storageMethods_t* ansmethods =
        (storageMethods_t*) R_ExternalPtrAddr(sansmethods);
    getExtElt_t getElt = xmethods->complexGetElt;
    setExtElt_t setElt = ansmethods->setExtElt;
    int i;
    int n = asInteger(GET_SLOT(GET_SLOT(x, install("storage")),
                               install("length")));
    SEXP xstorage = GET_STORAGE_EXT_PTR(GET_SLOT(x, install("storage")));
    SEXP ansStorage = GET_STORAGE_EXT_PTR(GET_SLOT(ans, install("storage")));
    primitiveElt_t elt;
    int warn = 0, iop;

    genericName = STRING_ELT(genericName, 0);
    iop = genericName2opcode(CHAR(genericName), LENGTH(genericName));
    if (iop == CUMSUM_GENERIC || iop == CUMPROD_GENERIC) {
        double accum_r;
        double accum_i;

        if (iop == CUMSUM_GENERIC) { /* cumsum */
            accum_r = accum_i = 0.0;
            for (i = 0; i < n; i++) {
                getElt(xstorage, xmethods, i, &elt, &warn);
                if (ISNAN(elt.cmplx.r) && ISNAN(elt.cmplx.i))
                    break;
                if (ISNAN(elt.cmplx.r)) {
                    elt.cmplx.i = elt.cmplx.r;
                    break;
                }
                if (ISNAN(elt.cmplx.i)) {
                    elt.cmplx.r = elt.cmplx.i;
                    break;
                }
                elt.cmplx.r = accum_r += elt.cmplx.r;
                elt.cmplx.i = accum_i += elt.cmplx.i;
                setElt(ansStorage, ansmethods, i, &elt, &warn);
            }
        } else { /* cumprod */
            accum_r = 1.0;
            accum_i = 0.0;
            for (i = 0; i < n; i++) {
                double tmp;
                getElt(xstorage, xmethods, i, &elt, &warn);
                if (ISNAN(elt.cmplx.r) && ISNAN(elt.cmplx.i))
                    break;
                if (ISNAN(elt.cmplx.r)) {
                    elt.cmplx.i = elt.cmplx.r;
                    break;
                }
                if (ISNAN(elt.cmplx.i)) {
                    elt.cmplx.r = elt.cmplx.i;
                    break;
                }
                tmp = accum_r;
                accum_r = accum_r*elt.cmplx.r-accum_i*elt.cmplx.i;
                elt.cmplx.i = accum_i =
                    tmp*elt.cmplx.i+accum_i*elt.cmplx.r;
                elt.cmplx.r = accum_r;
                setElt(ansStorage, ansmethods, i, &elt, &warn);
            }
        }

        while (i < n) {
            setElt(ansStorage, ansmethods, i++, &elt, &warn);
        }
    } else {
        void (*f)(Rcomplex*);

        switch (iop) {
        case LOG_GENERIC:
            f = &z_log;
            break;
        case SQRT_GENERIC:
            f = &z_sqrt;
            break;
        case LOG10_GENERIC:
            f = &z_log10;
            break;
        case ACOS_GENERIC:
            f = &z_acos;
            break;
        case ACOSH_GENERIC:
            f = &z_acosh;
            break;
        case ASIN_GENERIC:
            f = &z_asin;
            break;
        case ASINH_GENERIC:
            f = &z_asinh;
            break;
        case ATAN_GENERIC:
            f = &z_atan;
            break;
        case ATANH_GENERIC:
            f = &z_atanh;
            break;
        case COS_GENERIC:
            f = &z_cos;
            break;
        case COSH_GENERIC:
            f = &z_cosh;
            break;
        case EXP_GENERIC:
            f = &z_exp;
            break;
        case SIN_GENERIC:
            f = &z_sin;
            break;
        case SINH_GENERIC:
            f = &z_sinh;
            break;
        case TAN_GENERIC:
            f = &z_tan;
            break;
        case TANH_GENERIC:
            f = &z_tanh;
            break;
        default:
            error("unimplemented complex operation");
            return R_NilValue; /* -Wall */
        }
        for (i = 0; i < n; i++) {
            getElt(xstorage, xmethods, i, &elt, &warn);
            if (ISNAN(elt.numeric))
                setElt(ansStorage, ansmethods, i, &elt, &warn);
            else {
                f(&(elt.cmplx));
                setElt(ansStorage, ansmethods, i, &elt, &warn);
            }
        }
    }
    return ans;
}

static SEXP
externalVectorComplex(SEXP genericName, SEXP x, SEXP sxmethods,
                      SEXP ans, SEXP sansmethods)
{
    storageMethods_t* xmethods =
        (storageMethods_t*) R_ExternalPtrAddr(sxmethods);
    storageMethods_t* ansmethods =
        (storageMethods_t*) R_ExternalPtrAddr(sansmethods);
    setExtElt_t setElt = ansmethods->setExtElt;
    int i;
    int n = asInteger(GET_SLOT(GET_SLOT(x, install("storage")),
                               install("length")));
    SEXP xstorage = GET_STORAGE_EXT_PTR(GET_SLOT(x, install("storage")));
    SEXP ansStorage = GET_STORAGE_EXT_PTR(GET_SLOT(ans, install("storage")));
    primitiveElt_t elt;
    int warn = 0, iop;

    genericName = STRING_ELT(genericName, 0);
    iop = genericName2opcode(CHAR(genericName), LENGTH(genericName));
    if (xmethods->mode == CPLXSXP) {
        getExtElt_t getElt = xmethods->complexGetElt;

        switch (iop) {
        case ARG_GENERIC:
            for (i = 0; i < n; i++) {
                getElt(xstorage, xmethods, i, &elt, &warn);
#ifdef IEEE_754
                elt.numeric = atan2(elt.cmplx.i, elt.cmplx.r);
#else
                if (ISNAN(elt.cmplx.r) || ISNAN(elt.cmplx.i))
                    elt.numeric = NA_REAL;
                else elt.numeric = atan2(elt.cmplx.i, elt.cmplx.r);
#endif
                setElt(ansStorage, ansmethods, i, &elt, &warn);
            }
            break;
        case CONJ_GENERIC:
            for (i = 0; i < n; i++) {
                getElt(xstorage, xmethods, i, &elt, &warn);
#ifdef IEEE_754
                elt.cmplx.i = -elt.cmplx.i;
#else
                if (ISNAN(elt.cmplx.r) || ISNAN(elt.cmplx.i)) {
                    elt.cmplx.r = NA_REAL;
                    elt.cmplx.i = NA_REAL;
                } else elt.cmplx.i = -elt.cmplx.i;
#endif
                setElt(ansStorage, ansmethods, i, &elt, &warn);
            }
            break;
        case IM_GENERIC:
            for (i = 0; i < n; i++) {
                getElt(xstorage, xmethods, i, &elt, &warn);
                elt.numeric = elt.cmplx.i;
                setElt(ansStorage, ansmethods, i, &elt, &warn);
            }
            break;
        case RE_GENERIC:
            for (i = 0; i < n; i++) {
                getElt(xstorage, xmethods, i, &elt, &warn);
                elt.numeric = elt.cmplx.r;
                setElt(ansStorage, ansmethods, i, &elt, &warn);
            }
            break;
        case ABS_GENERIC:
            for (i = 0; i < n; i++) {
                getElt(xstorage, xmethods, i, &elt, &warn);
#ifdef IEEE_754
                elt.numeric = hypot(elt.cmplx.r, elt.cmplx.i);
#else
                if (ISNAN(elt.cmplx.r) || ISNAN(elt.cmplx.i))
                    elt.numeric = NA_REAL;
                else elt.numeric = hypot(elt.cmplx.r, elt.cmplx.i);
#endif
                setElt(ansStorage, ansmethods, i, &elt, &warn);
            }
            break;
        }
    } else {
        getExtElt_t getElt = xmethods->numericGetElt;

        switch (iop) {
        case CONJ_GENERIC:
        case RE_GENERIC:
            for (i = 0; i < n; i++) {
                getElt(xstorage, xmethods, i, &elt, &warn);
                setElt(ansStorage, ansmethods, i, &elt, &warn);
            }
            break;
        case ARG_GENERIC:
        case IM_GENERIC:
            for (i = 0; i < n; i++) {
                getElt(xstorage, xmethods, i, &elt, &warn);
                if (!ISNAN(elt.numeric))
                    elt.numeric = 0.0;
                setElt(ansStorage, ansmethods, i, &elt, &warn);
            }
            break;
        case ABS_GENERIC:
            for (i = 0; i < n; i++) {
                getElt(xstorage, xmethods, i, &elt, &warn);
#ifdef IEEE_754
                elt.numeric = fabs(elt.numeric);
#else
                if (!ISNAN(elt.numeric))
                    elt.numeric = fabs(elt.numeric);
#endif
                setElt(ansStorage, ansmethods, i, &elt, &warn);
            }
            break;
        }
    }
    return ans;
}

#define USES_TWOS_COMPLEMENT 1

#if USES_TWOS_COMPLEMENT
# define OPPOSITE_SIGNS(x, y) ((x < 0) ^ (y < 0))
# define GOODISUM(x, y, z) (((x) > 0) ? ((y) < (z)) : ! ((y) < (z)))
# define GOODIDIFF(x, y, z) (!(OPPOSITE_SIGNS(x, y) && OPPOSITE_SIGNS(x, z)))
#else
# define GOODISUM(x, y, z) ((double) (x) + (double) (y) == (z))
# define GOODIDIFF(x, y, z) ((double) (x) - (double) (y) == (z))
#endif
#define GOODIPROD(x, y, z) ((double) (x) * (double) (y) == (z))
#define INTEGER_OVERFLOW_WARNING "NAs produced by integer overflow"

static void
integerBinaryArith(int iop,
                   int nx, SEXP xstorage, storageMethods_t* xmethods,
                   int ny, SEXP ystorage, storageMethods_t* ymethods,
                   int n, SEXP ansStorage, storageMethods_t* ansmethods)
{
    getExtElt_t xgetElt = xmethods->integerGetElt;
    getExtElt_t ygetElt = ymethods->integerGetElt;
    setExtElt_t setElt = ansmethods->setExtElt;
    primitiveElt_t xelt;
    primitiveElt_t yelt;
    int xwarn;
    int ywarn;
    int warn;
    int i, i1, i2;
    int naflag = FALSE;

    xwarn = ywarn = warn = 0;
    switch (iop) {
    case PLUS_GENERIC:
        mod_iterate(nx, ny, i1, i2) {
            xgetElt(xstorage, xmethods, i1, &xelt, &xwarn);
            ygetElt(ystorage, ymethods, i2, &yelt, &ywarn);
            if (xelt.integer != NA_INTEGER) {
                if (yelt.numeric == NA_INTEGER)
                    xelt.numeric = NA_INTEGER;
                else {
                    int val = xelt.integer + yelt.integer;
                    if (val != NA_INTEGER && GOODISUM(xelt.integer,
                                                      yelt.integer, val))
                        xelt.integer = val;
                    else {
                        xelt.integer = NA_INTEGER;
                        naflag = TRUE;
                    }
                }
            }
            setElt(ansStorage, ansmethods, i, &xelt, &warn);
        }
        if (naflag)
            warning(INTEGER_OVERFLOW_WARNING);
        break;
    case MINUS_GENERIC:
        mod_iterate(nx, ny, i1, i2) {
            xgetElt(xstorage, xmethods, i1, &xelt, &xwarn);
            ygetElt(ystorage, ymethods, i2, &yelt, &ywarn);
            if (xelt.integer != NA_INTEGER) {
                if (yelt.numeric == NA_INTEGER)
                    xelt.numeric = NA_INTEGER;
                else {
                    int val = xelt.integer - yelt.integer;
                    if (val != NA_INTEGER && GOODIDIFF(xelt.integer,
                                                       yelt.integer, val))
                        xelt.integer = val;
                    else {
                        xelt.integer = NA_INTEGER;
                        naflag = TRUE;
                    }
                }
            }
            setElt(ansStorage, ansmethods, i, &xelt, &warn);
        }
        if (naflag)
            warning(INTEGER_OVERFLOW_WARNING);
        break;
    case MULT_GENERIC:
        mod_iterate(nx, ny, i1, i2) {
            xgetElt(xstorage, xmethods, i1, &xelt, &xwarn);
            ygetElt(ystorage, ymethods, i2, &yelt, &ywarn);
            if (xelt.integer != NA_INTEGER) {
                if (yelt.numeric == NA_INTEGER)
                    xelt.numeric = NA_INTEGER;
                else {
                    int val = xelt.integer * yelt.integer;
                    if (val != NA_INTEGER && GOODIPROD(xelt.integer,
                                                       yelt.integer, val))
                        xelt.integer = val;
                    else {
                        xelt.integer = NA_INTEGER;
                        naflag = TRUE;
                    }
                }
            }
            setElt(ansStorage, ansmethods, i, &xelt, &warn);
        }
        if (naflag)
            warning(INTEGER_OVERFLOW_WARNING);
        break;
    case DIV_GENERIC:
        mod_iterate(nx, ny, i1, i2) {
            xgetElt(xstorage, xmethods, i1, &xelt, &xwarn);
            ygetElt(ystorage, ymethods, i2, &yelt, &ywarn);
#ifdef IEEE_754
            if (xelt.integer == NA_INTEGER || yelt.integer == NA_INTEGER)
#else
            if (xelt.integer == NA_INTEGER || yelt.integer == NA_INTEGER ||
                yelt.integer == 0)
#endif
                xelt.numeric = NA_REAL;
            else xelt.numeric = (double) xelt.integer/(double) yelt.integer;
            setElt(ansStorage, ansmethods, i, &xelt, &warn);
        }
        break;
    case POW_GENERIC:
        mod_iterate(nx, ny, i1, i2) {
            xgetElt(xstorage, xmethods, i1, &xelt, &xwarn);
            ygetElt(ystorage, ymethods, i2, &yelt, &ywarn);
            if (xelt.integer == NA_INTEGER || yelt.integer == NA_INTEGER)
                xelt.numeric = NA_REAL;
            else xelt.numeric = MATH_CHECK(R_pow((double) xelt.integer,
                                                 (double) yelt.integer));
            setElt(ansStorage, ansmethods, i, &xelt, &warn);
        }
        break;
    case INTMOD_GENERIC:
        mod_iterate(nx, ny, i1, i2) {
            xgetElt(xstorage, xmethods, i1, &xelt, &xwarn);
            ygetElt(ystorage, ymethods, i2, &yelt, &ywarn);
            if (xelt.integer != NA_INTEGER) {
                if (yelt.integer == NA_INTEGER)
                    xelt.integer = NA_INTEGER;
                else if (xelt.integer >= 0 && yelt.integer >= 0) {
                    xelt.integer = xelt.integer % yelt.integer;
                } else {
                    double tmpx = (double) xelt.integer;
                    double tmpy = (double) yelt.integer;
                    xelt.integer = myfmod(tmpx, tmpy);
                }
            }
            setElt(ansStorage, ansmethods, i, &xelt, &warn);
        }
        break;
    case INTDIV_GENERIC:
        mod_iterate(nx, ny, i1, i2) {
            xgetElt(xstorage, xmethods, i1, &xelt, &xwarn);
            ygetElt(ystorage, ymethods, i2, &yelt, &ywarn);
            if (xelt.integer != NA_INTEGER) {
                if (yelt.integer == NA_INTEGER)
                    xelt.integer = NA_INTEGER;
                else if (yelt.integer == 0)
                    xelt.integer = 0;
                else xelt.integer = floor((double) xelt.integer/
                                          (double) yelt.integer);
            }
            setElt(ansStorage, ansmethods, i, &xelt, &warn);
        }
        break;
    }
    CoercionWarning(xwarn);
    CoercionWarning(ywarn);
    CoercionWarning(warn);
}

static void
numericBinaryArith(int iop,
                   int nx, SEXP xstorage, storageMethods_t* xmethods,
                   int ny, SEXP ystorage, storageMethods_t* ymethods,
                   int n, SEXP ansStorage, storageMethods_t* ansmethods)
{
    getExtElt_t xgetElt = xmethods->numericGetElt;
    getExtElt_t ygetElt = ymethods->numericGetElt;
    setExtElt_t setElt = ansmethods->setExtElt;
    primitiveElt_t xelt;
    primitiveElt_t yelt;
    int xwarn;
    int ywarn;
    int warn;
    int i, i1, i2;

    xwarn = ywarn = warn = 0;
    switch (iop) {
    case PLUS_GENERIC:
        mod_iterate(nx, ny, i1, i2) {
            xgetElt(xstorage, xmethods, i1, &xelt, &xwarn);
            ygetElt(ystorage, ymethods, i2, &yelt, &ywarn);
#ifdef IEEE_754
            xelt.numeric += yelt.numeric;
#else
            if (!ISNA(xelt.numeric)) {
                if (ISNA(yelt.numeric))
                    xelt.numeric = NA_REAL;
                else xelt.numeric = MATH_CHECK(xelt.numeric + yelt.numeric);
            }
#endif
            setElt(ansStorage, ansmethods, i, &xelt, &warn);
        }
        break;
    case MINUS_GENERIC:
        mod_iterate(nx, ny, i1, i2) {
            xgetElt(xstorage, xmethods, i1, &xelt, &xwarn);
            ygetElt(ystorage, ymethods, i2, &yelt, &ywarn);
#ifdef IEEE_754
            xelt.numeric -= yelt.numeric;
#else
            if (!ISNA(xelt.numeric)) {
                if (ISNA(yelt.numeric))
                    xelt.numeric = NA_REAL;
                else xelt.numeric = MATH_CHECK(xelt.numeric - yelt.numeric);
            }
#endif
            setElt(ansStorage, ansmethods, i, &xelt, &warn);
        }
        break;
    case MULT_GENERIC:
        mod_iterate(nx, ny, i1, i2) {
            xgetElt(xstorage, xmethods, i1, &xelt, &xwarn);
            ygetElt(ystorage, ymethods, i2, &yelt, &ywarn);
#ifdef IEEE_754
            xelt.numeric *= yelt.numeric;
#else
            if (!ISNA(xelt.numeric)) {
                if (ISNA(yelt.numeric))
                    xelt.numeric = NA_REAL;
                else xelt.numeric = MATH_CHECK(xelt.numeric * yelt.numeric);
            }
#endif
            setElt(ansStorage, ansmethods, i, &xelt, &warn);
        }
        break;
    case DIV_GENERIC:
        mod_iterate(nx, ny, i1, i2) {
            xgetElt(xstorage, xmethods, i1, &xelt, &xwarn);
            ygetElt(ystorage, ymethods, i2, &yelt, &ywarn);
#ifdef IEEE_754
            xelt.numeric /= yelt.numeric;
#else
            if (!ISNA(xelt.numeric)) {
                if (ISNA(yelt.numeric))
                    xelt.numeric = NA_REAL;
                else if (yelt.numeric == 0.0)
                    xelt.numeric = NA_REAL;
                else xelt.numeric = MATH_CHECK(xelt.numeric + yelt.numeric);
            }
#endif
            setElt(ansStorage, ansmethods, i, &xelt, &warn);
        }
        break;
    case POW_GENERIC:
        mod_iterate(nx, ny, i1, i2) {
            xgetElt(xstorage, xmethods, i1, &xelt, &xwarn);
            ygetElt(ystorage, ymethods, i2, &yelt, &ywarn);
#ifdef IEEE_754
            xelt.numeric = R_pow(xelt.numeric, yelt.numeric);
#else
            if (!ISNA(xelt.numeric)) {
                if (ISNA(yelt.numeric))
                    xelt.numeric = NA_REAL;
                else xelt.numeric = MATH_CHECK(R_pow(xelt.numeric,
                                                     yelt.numeric));
            }
#endif
            setElt(ansStorage, ansmethods, i, &xelt, &warn);
        }
        break;
    case INTMOD_GENERIC:
        mod_iterate(nx, ny, i1, i2) {
            xgetElt(xstorage, xmethods, i1, &xelt, &xwarn);
            ygetElt(ystorage, ymethods, i2, &yelt, &ywarn);
#ifdef IEEE_754
            xelt.numeric = myfmod(xelt.numeric, yelt.numeric);
#else
            if (!ISNA(xelt.numeric)) {
                if (ISNA(yelt.numeric))
                    xelt.numeric = NA_REAL;
                else xelt.numeric = MATH_CHECK(myfmod(xelt.numeric,
                                                      yelt.numeric));
            }
#endif
            setElt(ansStorage, ansmethods, i, &xelt, &warn);
        }
        break;
    case INTDIV_GENERIC:
        mod_iterate(nx, ny, i1, i2) {
            xgetElt(xstorage, xmethods, i1, &xelt, &xwarn);
            ygetElt(ystorage, ymethods, i2, &yelt, &ywarn);
#ifdef IEEE_754
            xelt.numeric = floor(xelt.numeric/yelt.numeric);
#else
            if (!ISNA(xelt.numeric)) {
                if (ISNA(yelt.numeric))
                    xelt.numeric = NA_REAL;
                else xelt.numeric = MATH_CHECK(floor(xelt.numeric/
                                                     yelt.numeric));
            }
#endif
            setElt(ansStorage, ansmethods, i, &xelt, &warn);
        }
        break;
    }
    CoercionWarning(xwarn);
    CoercionWarning(ywarn);
    CoercionWarning(warn);
}

static void
complex_div(primitiveElt_t* a, primitiveElt_t* b)
{
    double ratio, den;
    double abr, abi;
    double tmp;

    if( (abr = b->cmplx.r) < 0)
        abr = - abr;
    if( (abi = b->cmplx.i) < 0)
	abi = - abi;
    if( abr <= abi ) {
#ifndef IEEE_754
	if(abi == 0) {
	    a->cmplx.r = NA_REAL;
	    a->cmplx.i = NA_REAL;
	    return;
	}
#endif
	ratio = b->cmplx.r / b->cmplx.i ;
	den = b->cmplx.i * (1 + ratio*ratio);
    tmp = a->cmplx.r;
	a->cmplx.r = (tmp*ratio + a->cmplx.i) / den;
	a->cmplx.i = (a->cmplx.i*ratio - tmp) / den;
    }
    else {
	ratio = b->cmplx.i / b->cmplx.r ;
	den = b->cmplx.r * (1 + ratio*ratio);
    tmp = a->cmplx.r;
	a->cmplx.r = (tmp + a->cmplx.i*ratio) / den;
	a->cmplx.i = (a->cmplx.i - tmp*ratio) / den;
    }
}

static void
complex_pow(primitiveElt_t* a, primitiveElt_t* b)
{
    double logr, logi, x, y;
    int ib;

    if(b->cmplx.i == 0.) {		/* ^ "real" : be fast (and more accurate)*/
	if(b->cmplx.r == 1.) {	/* a^1 */
	    return;
	}
	if(a->cmplx.i == 0. && a->cmplx.r >= 0.) {
	    a->cmplx.r = R_pow(a->cmplx.r, b->cmplx.r); return;
	}
	if(a->cmplx.r == 0. &&
       b->cmplx.r == (ib = (int)b->cmplx.r)) { /* (|a|*i)^b */
	    x = R_pow_di(a->cmplx.i, ib);
	    if(ib % 2) {	/* ib is odd ==> imaginary r */
		a->cmplx.i = ((ib>0 && ib %4 == 3)||(ib<0 && (-ib)%4 == 1))? -x : x;
	    } else {		/* even exponent b : real r */
		a->cmplx.r = (ib %4)? -x : x; a->cmplx.i = 0.;
	    }
	    return;
	}
    }
    logr = log(hypot(a->cmplx.r, a->cmplx.i) );
    logi = atan2(a->cmplx.i, a->cmplx.r);
    x = exp( logr * b->cmplx.r - logi * b->cmplx.i );
    y = logr * b->cmplx.i + logi * b->cmplx.r;
    a->cmplx.r = x * cos(y);
    a->cmplx.i = x * sin(y);
}

static void
complexBinaryArith(int iop,
                   int nx, SEXP xstorage, storageMethods_t* xmethods,
                   int ny, SEXP ystorage, storageMethods_t* ymethods,
                   int n, SEXP ansStorage, storageMethods_t* ansmethods)
{
    getExtElt_t xgetElt = xmethods->numericGetElt;
    getExtElt_t ygetElt = ymethods->numericGetElt;
    setExtElt_t setElt = ansmethods->setExtElt;
    primitiveElt_t xelt;
    primitiveElt_t yelt;
    int xwarn;
    int ywarn;
    int warn;
    int i, i1, i2;

    xwarn = ywarn = warn = 0;
    switch (iop) {
    case PLUS_GENERIC:
        mod_iterate(nx, ny, i1, i2) {
            xgetElt(xstorage, xmethods, i1, &xelt, &xwarn);
            ygetElt(ystorage, ymethods, i2, &yelt, &ywarn);
#ifdef IEEE_754
            xelt.cmplx.r += yelt.cmplx.r;
            xelt.cmplx.i += yelt.cmplx.i;
#else
            if (ISNAN(xelt.cmplx.r) || ISNAN(xelt.cmplx.i) ||
                ISNAN(yelt.cmplx.r) || ISNAN(yelt.cmplx.i)) {
                xelt.cmplx.r = NA_REAL;
                xelt.cmplx.i = NA_REAL;
            } else {
                xelt.cmplx.r = MATH_CHECK(xelt.cmplx.r + yelt.cmplx.r);
                xelt.cmplx.i = MATH_CHECK(xelt.cmplx.i + yelt.cmplx.i);
            }
#endif
            setElt(ansStorage, ansmethods, i, &xelt, &warn);
        }
        break;
    case MINUS_GENERIC:
        mod_iterate(nx, ny, i1, i2) {
            xgetElt(xstorage, xmethods, i1, &xelt, &xwarn);
            ygetElt(ystorage, ymethods, i2, &yelt, &ywarn);
#ifdef IEEE_754
            xelt.cmplx.r -= yelt.cmplx.r;
            xelt.cmplx.i -= yelt.cmplx.i;
#else
            if (ISNAN(xelt.cmplx.r) || ISNAN(xelt.cmplx.i) ||
                ISNAN(yelt.cmplx.r) || ISNAN(yelt.cmplx.i)) {
                xelt.cmplx.r = NA_REAL;
                xelt.cmplx.i = NA_REAL;
            } else {
                xelt.cmplx.r = MATH_CHECK(xelt.cmplx.r - yelt.cmplx.r);
                xelt.cmplx.i = MATH_CHECK(xelt.cmplx.i - yelt.cmplx.i);
            }
#endif
            setElt(ansStorage, ansmethods, i, &xelt, &warn);
        }
        break;
    case MULT_GENERIC:
        mod_iterate(nx, ny, i1, i2) {
            double tmp;
            xgetElt(xstorage, xmethods, i1, &xelt, &xwarn);
            ygetElt(ystorage, ymethods, i2, &yelt, &ywarn);
#ifdef IEEE_754
            tmp = xelt.cmplx.r;
            xelt.cmplx.r = tmp*yelt.cmplx.r - xelt.cmplx.i*yelt.cmplx.i;
            xelt.cmplx.i = tmp*yelt.cmplx.i + xelt.cmplx.i*yelt.cmplx.r;
#else
            if (ISNAN(xelt.cmplx.r) || ISNAN(xelt.cmplx.i) ||
                ISNAN(yelt.cmplx.r) || ISNAN(yelt.cmplx.i)) {
                xelt.cmplx.r = NA_REAL;
                xelt.cmplx.i = NA_REAL;
            } else {
                xelt.cmplx.r = MATH_CHECK(tmp*yelt.cmplx.r -
                                          xelt.cmplx.i*yelt.cmplx.i);
                xelt.cmplx.i = MATH_CHECK(tmp*yelt.cmplx.i +
                                          xelt.cmplx.i*yelt.cmplx.r);
            }
#endif
            setElt(ansStorage, ansmethods, i, &xelt, &warn);
        }
        break;
    case DIV_GENERIC:
        mod_iterate(nx, ny, i1, i2) {
            xgetElt(xstorage, xmethods, i1, &xelt, &xwarn);
            ygetElt(ystorage, ymethods, i2, &yelt, &ywarn);
#ifdef IEEE_754
            complex_div(&xelt, &yelt);
#else
            if (ISNAN(xelt.cmplx.r) || ISNAN(xelt.cmplx.i) ||
                ISNAN(yelt.cmplx.r) || ISNAN(yelt.cmplx.i)) {
                xelt.cmplx.r = NA_REAL;
                xelt.cmplx.i = NA_REAL;
            } else {
                complex_div(&xelt, &yelt);
            }
#endif
            setElt(ansStorage, ansmethods, i, &xelt, &warn);
        }
        break;
    case POW_GENERIC:
        mod_iterate(nx, ny, i1, i2) {
            xgetElt(xstorage, xmethods, i1, &xelt, &xwarn);
            ygetElt(ystorage, ymethods, i2, &yelt, &ywarn);
#ifdef IEEE_754
            complex_pow(&xelt, &yelt);
#else
            if (ISNAN(xelt.cmplx.r) || ISNAN(xelt.cmplx.i) ||
                ISNAN(yelt.cmplx.r) || ISNAN(yelt.cmplx.i)) {
                xelt.cmplx.r = NA_REAL;
                xelt.cmplx.i = NA_REAL;
            } else {
                complex_pow(&xelt, &yelt);
            }
#endif
            setElt(ansStorage, ansmethods, i, &xelt, &warn);
        }
        break;
    default:
        error("unimplemented complex operation");
    }
    CoercionWarning(xwarn);
    CoercionWarning(ywarn);
    CoercionWarning(warn);
}

static SEXP
externalVectorBinaryArith(SEXP genericName, SEXP x, SEXP sxmethods,
                          SEXP y, SEXP symethods)
{
    SEXP xstorage = GET_SLOT(x, install("storage"));
    int nx = asInteger(GET_SLOT(xstorage, install("length")));
    SEXP ystorage = GET_SLOT(y, install("storage"));
    int ny = asInteger(GET_SLOT(ystorage, install("length")));
    storageMethods_t* xmethods =
        (storageMethods_t*) R_ExternalPtrAddr(sxmethods);
    storageMethods_t* ymethods =
        (storageMethods_t*) R_ExternalPtrAddr(symethods);
    int mode;
    int n, iop;
    SEXP ans, ansStorage;
    storageMethods_t* ansmethods;

    genericName = STRING_ELT(genericName, 0);
    iop = genericName2opcode(CHAR(genericName), LENGTH(genericName));

    if (xmethods->mode == CPLXSXP || ymethods->mode == CPLXSXP)
        mode = CPLXSXP;
    else if (xmethods->mode == REALSXP || ymethods->mode == REALSXP)
        mode = REALSXP;
    else mode = INTSXP;
    if (nx == 0 || ny == 0)
        return allocVector(mode, 0);
    n = (nx < ny)?ny:nx;
    ans = getArithmeticAnswerStorage(xstorage, xmethods, nx,
                                     ystorage, ymethods, ny,
                                     (mode == INTSXP &&
                                      (iop == DIV_GENERIC ||
                                       iop == POW_GENERIC))?REALSXP:mode);
    PROTECT(ans);
    xstorage = GET_STORAGE_EXT_PTR(xstorage);
    ystorage = GET_STORAGE_EXT_PTR(ystorage);
    ansStorage = GET_STORAGE_EXT_PTR(ans);
    ansmethods = getStorageMethods(ans);
    switch (mode) {
    case CPLXSXP:
        complexBinaryArith(iop, nx, xstorage, xmethods,
                           ny, ystorage, ymethods,
                           n, ansStorage, ansmethods);
        break;
    case REALSXP:
        numericBinaryArith(iop, nx, xstorage, xmethods,
                           ny, ystorage, ymethods,
                           n, ansStorage, ansmethods);
        break;
    case INTSXP:
        integerBinaryArith(iop, nx, xstorage, xmethods,
                           ny, ystorage, ymethods,
                           n, ansStorage, ansmethods);
        break;
    }
    UNPROTECT(1);
    return ans;
}

static void
integerCompare(int iop,
               int nx, SEXP xstorage, storageMethods_t* xmethods,
               int ny, SEXP ystorage, storageMethods_t* ymethods,
               int n, SEXP ansStorage, storageMethods_t* ansmethods)
{
    getExtElt_t xgetElt = xmethods->integerGetElt;
    getExtElt_t ygetElt = ymethods->integerGetElt;
    setExtElt_t setElt = ansmethods->setExtElt;
    primitiveElt_t xelt;
    primitiveElt_t yelt;
    int xwarn;
    int ywarn;
    int warn;
    int i, i1, i2;

    xwarn = ywarn = warn = 0;
    switch (iop) {
    case EQ_GENERIC:
        mod_iterate(nx, ny, i1, i2) {
            xgetElt(xstorage, xmethods, i1, &xelt, &xwarn);
            ygetElt(ystorage, ymethods, i2, &yelt, &ywarn);
            if (xelt.integer == NA_INTEGER || yelt.integer == NA_INTEGER)
                xelt.logical = NA_LOGICAL;
            else xelt.logical = (xelt.integer == yelt.integer);
            setElt(ansStorage, ansmethods, i, &xelt, &warn);
        }
        break;
    case NE_GENERIC:
        mod_iterate(nx, ny, i1, i2) {
            xgetElt(xstorage, xmethods, i1, &xelt, &xwarn);
            ygetElt(ystorage, ymethods, i2, &yelt, &ywarn);
            if (xelt.integer == NA_INTEGER || yelt.integer == NA_INTEGER)
                xelt.logical = NA_LOGICAL;
            else xelt.logical = (xelt.integer != yelt.integer);
            setElt(ansStorage, ansmethods, i, &xelt, &warn);
        }
        break;
    case LT_GENERIC:
        mod_iterate(nx, ny, i1, i2) {
            xgetElt(xstorage, xmethods, i1, &xelt, &xwarn);
            ygetElt(ystorage, ymethods, i2, &yelt, &ywarn);
            if (xelt.integer == NA_INTEGER || yelt.integer == NA_INTEGER)
                xelt.logical = NA_LOGICAL;
            else xelt.logical = (xelt.integer < yelt.integer);
            setElt(ansStorage, ansmethods, i, &xelt, &warn);
        }
        break;
    case GT_GENERIC:
        mod_iterate(nx, ny, i1, i2) {
            xgetElt(xstorage, xmethods, i1, &xelt, &xwarn);
            ygetElt(ystorage, ymethods, i2, &yelt, &ywarn);
            if (xelt.integer == NA_INTEGER || yelt.integer == NA_INTEGER)
                xelt.logical = NA_LOGICAL;
            else xelt.logical = (xelt.integer > yelt.integer);
            setElt(ansStorage, ansmethods, i, &xelt, &warn);
        }
        break;
    case LE_GENERIC:
        mod_iterate(nx, ny, i1, i2) {
            xgetElt(xstorage, xmethods, i1, &xelt, &xwarn);
            ygetElt(ystorage, ymethods, i2, &yelt, &ywarn);
            if (xelt.integer == NA_INTEGER || yelt.integer == NA_INTEGER)
                xelt.logical = NA_LOGICAL;
            else xelt.logical = (xelt.integer <= yelt.integer);
            setElt(ansStorage, ansmethods, i, &xelt, &warn);
        }
        break;
    case GE_GENERIC:
        mod_iterate(nx, ny, i1, i2) {
            xgetElt(xstorage, xmethods, i1, &xelt, &xwarn);
            ygetElt(ystorage, ymethods, i2, &yelt, &ywarn);
            if (xelt.integer == NA_INTEGER || yelt.integer == NA_INTEGER)
                xelt.logical = NA_LOGICAL;
            else xelt.logical = (xelt.integer >= yelt.integer);
            setElt(ansStorage, ansmethods, i, &xelt, &warn);
        }
        break;
    }
    CoercionWarning(xwarn);
    CoercionWarning(ywarn);
    CoercionWarning(warn);
}

static void
numericCompare(int iop,
               int nx, SEXP xstorage, storageMethods_t* xmethods,
               int ny, SEXP ystorage, storageMethods_t* ymethods,
               int n, SEXP ansStorage, storageMethods_t* ansmethods)
{
    getExtElt_t xgetElt = xmethods->numericGetElt;
    getExtElt_t ygetElt = ymethods->numericGetElt;
    setExtElt_t setElt = ansmethods->setExtElt;
    primitiveElt_t xelt;
    primitiveElt_t yelt;
    int xwarn;
    int ywarn;
    int warn;
    int i, i1, i2;

    xwarn = ywarn = warn = 0;
    switch (iop) {
    case EQ_GENERIC:
        mod_iterate(nx, ny, i1, i2) {
            xgetElt(xstorage, xmethods, i1, &xelt, &xwarn);
            ygetElt(ystorage, ymethods, i2, &yelt, &ywarn);
            if (ISNAN(xelt.numeric) || ISNAN(yelt.numeric))
                xelt.logical = NA_LOGICAL;
            else xelt.logical = (xelt.numeric == yelt.numeric);
            setElt(ansStorage, ansmethods, i, &xelt, &warn);
        }
        break;
    case NE_GENERIC:
        mod_iterate(nx, ny, i1, i2) {
            xgetElt(xstorage, xmethods, i1, &xelt, &xwarn);
            ygetElt(ystorage, ymethods, i2, &yelt, &ywarn);
            if (ISNAN(xelt.numeric) || ISNAN(yelt.numeric))
                xelt.logical = NA_LOGICAL;
            else xelt.logical = (xelt.numeric != yelt.numeric);
            setElt(ansStorage, ansmethods, i, &xelt, &warn);
        }
        break;
    case LT_GENERIC:
        mod_iterate(nx, ny, i1, i2) {
            xgetElt(xstorage, xmethods, i1, &xelt, &xwarn);
            ygetElt(ystorage, ymethods, i2, &yelt, &ywarn);
            if (ISNAN(xelt.numeric) || ISNAN(yelt.numeric))
                xelt.logical = NA_LOGICAL;
            else xelt.logical = (xelt.numeric < yelt.numeric);
            setElt(ansStorage, ansmethods, i, &xelt, &warn);
        }
        break;
    case GT_GENERIC:
        mod_iterate(nx, ny, i1, i2) {
            xgetElt(xstorage, xmethods, i1, &xelt, &xwarn);
            ygetElt(ystorage, ymethods, i2, &yelt, &ywarn);
            if (ISNAN(xelt.numeric) || ISNAN(yelt.numeric))
                xelt.logical = NA_LOGICAL;
            else xelt.logical = (xelt.numeric > yelt.numeric);
            setElt(ansStorage, ansmethods, i, &xelt, &warn);
        }
        break;
    case LE_GENERIC:
        mod_iterate(nx, ny, i1, i2) {
            xgetElt(xstorage, xmethods, i1, &xelt, &xwarn);
            ygetElt(ystorage, ymethods, i2, &yelt, &ywarn);
            if (ISNAN(xelt.numeric) || ISNAN(yelt.numeric))
                xelt.logical = NA_LOGICAL;
            else xelt.logical = (xelt.numeric <= yelt.numeric);
            setElt(ansStorage, ansmethods, i, &xelt, &warn);
        }
        break;
    case GE_GENERIC:
        mod_iterate(nx, ny, i1, i2) {
            xgetElt(xstorage, xmethods, i1, &xelt, &xwarn);
            ygetElt(ystorage, ymethods, i2, &yelt, &ywarn);
            if (ISNAN(xelt.numeric) || ISNAN(yelt.numeric))
                xelt.logical = NA_LOGICAL;
            else xelt.logical = (xelt.numeric >= yelt.numeric);
            setElt(ansStorage, ansmethods, i, &xelt, &warn);
        }
        break;
    }
    CoercionWarning(xwarn);
    CoercionWarning(ywarn);
    CoercionWarning(warn);
}

static void
complexCompare(int iop,
               int nx, SEXP xstorage, storageMethods_t* xmethods,
               int ny, SEXP ystorage, storageMethods_t* ymethods,
               int n, SEXP ansStorage, storageMethods_t* ansmethods)
{
    getExtElt_t xgetElt = xmethods->complexGetElt;
    getExtElt_t ygetElt = ymethods->complexGetElt;
    setExtElt_t setElt = ansmethods->setExtElt;
    primitiveElt_t xelt;
    primitiveElt_t yelt;
    int xwarn;
    int ywarn;
    int warn;
    int i, i1, i2;

    xwarn = ywarn = warn = 0;
    switch (iop) {
    case EQ_GENERIC:
        mod_iterate(nx, ny, i1, i2) {
            xgetElt(xstorage, xmethods, i1, &xelt, &xwarn);
            ygetElt(ystorage, ymethods, i2, &yelt, &ywarn);
            if (ISNAN(xelt.cmplx.r) || ISNAN(xelt.cmplx.i) ||
                ISNAN(yelt.cmplx.r) || ISNAN(yelt.cmplx.i))
                xelt.logical = NA_LOGICAL;
            else xelt.logical = (xelt.cmplx.r == yelt.cmplx.r &&
                                 xelt.cmplx.i == yelt.cmplx.i);
            setElt(ansStorage, ansmethods, i, &xelt, &warn);
        }
        break;
    case NE_GENERIC:
        mod_iterate(nx, ny, i1, i2) {
            xgetElt(xstorage, xmethods, i1, &xelt, &xwarn);
            ygetElt(ystorage, ymethods, i2, &yelt, &ywarn);
            if (ISNAN(xelt.cmplx.r) || ISNAN(xelt.cmplx.i) ||
                ISNAN(yelt.cmplx.r) || ISNAN(yelt.cmplx.i))
                xelt.logical = NA_LOGICAL;
            else xelt.logical = (xelt.cmplx.r != yelt.cmplx.r ||
                                 xelt.cmplx.i != yelt.cmplx.i);
            setElt(ansStorage, ansmethods, i, &xelt, &warn);
        }
        break;
    default:
        /* should not happen */
        error("ordering not defined for complex numbers");
        break;
    }
    CoercionWarning(xwarn);
    CoercionWarning(ywarn);
    CoercionWarning(warn);
}

#define STRCMP strcmp
#define streql(s, t)  (!strcmp((s), (t)))

static void
characterCompare(int iop,
                 int nx, SEXP xstorage, storageMethods_t* xmethods,
                 int ny, SEXP ystorage, storageMethods_t* ymethods,
                 int n, SEXP ansStorage, storageMethods_t* ansmethods)
{
    getExtElt_t xgetElt = xmethods->characterGetElt;
    getExtElt_t ygetElt = ymethods->characterGetElt;
    setExtElt_t setElt = ansmethods->setExtElt;
    primitiveElt_t xelt;
    primitiveElt_t yelt;
    int xwarn;
    int ywarn;
    int warn;
    int i, i1, i2;

    xwarn = ywarn = warn = 0;
    switch (iop) {
    case EQ_GENERIC:
        mod_iterate(nx, ny, i1, i2) {
            xgetElt(xstorage, xmethods, i1, &xelt, &xwarn);
            ygetElt(ystorage, ymethods, i2, &yelt, &ywarn);
            if (xelt.character == NA_STRING || yelt.character == NA_STRING)
                xelt.logical = NA_LOGICAL;
            else xelt.logical = strcmp(CHAR(xelt.character),
                                       CHAR(yelt.character)) == 0;
            setElt(ansStorage, ansmethods, i, &xelt, &warn);
        }
        break;
    case NE_GENERIC:
        mod_iterate(nx, ny, i1, i2) {
            xgetElt(xstorage, xmethods, i1, &xelt, &xwarn);
            ygetElt(ystorage, ymethods, i2, &yelt, &ywarn);
            if (xelt.character == NA_STRING || yelt.character == NA_STRING)
                xelt.logical = NA_LOGICAL;
            else xelt.logical = strcmp(CHAR(xelt.character),
                                       CHAR(yelt.character)) != 0;
            setElt(ansStorage, ansmethods, i, &xelt, &warn);
        }
        break;
    case LT_GENERIC:
        mod_iterate(nx, ny, i1, i2) {
            xgetElt(xstorage, xmethods, i1, &xelt, &xwarn);
            ygetElt(ystorage, ymethods, i2, &yelt, &ywarn);
            if (xelt.character == NA_STRING || yelt.character == NA_STRING)
                xelt.logical = NA_LOGICAL;
            else xelt.logical = STRCMP(CHAR(xelt.character),
                                       CHAR(yelt.character)) < 0;
            setElt(ansStorage, ansmethods, i, &xelt, &warn);
        }
        break;
    case GT_GENERIC:
        mod_iterate(nx, ny, i1, i2) {
            xgetElt(xstorage, xmethods, i1, &xelt, &xwarn);
            ygetElt(ystorage, ymethods, i2, &yelt, &ywarn);
            if (xelt.character == NA_STRING || yelt.character == NA_STRING)
                xelt.logical = NA_LOGICAL;
            else xelt.logical = STRCMP(CHAR(xelt.character),
                                       CHAR(yelt.character)) > 0;
            setElt(ansStorage, ansmethods, i, &xelt, &warn);
        }
        break;
    case LE_GENERIC:
        mod_iterate(nx, ny, i1, i2) {
            xgetElt(xstorage, xmethods, i1, &xelt, &xwarn);
            ygetElt(ystorage, ymethods, i2, &yelt, &ywarn);
            if (xelt.character == NA_STRING || yelt.character == NA_STRING)
                xelt.logical = NA_LOGICAL;
            else xelt.logical = STRCMP(CHAR(xelt.character),
                                       CHAR(yelt.character)) <= 0;
            setElt(ansStorage, ansmethods, i, &xelt, &warn);
        }
        break;
    case GE_GENERIC:
        mod_iterate(nx, ny, i1, i2) {
            xgetElt(xstorage, xmethods, i1, &xelt, &xwarn);
            ygetElt(ystorage, ymethods, i2, &yelt, &ywarn);
            if (xelt.character == NA_STRING || yelt.character == NA_STRING)
                xelt.logical = NA_LOGICAL;
            else xelt.logical = STRCMP(CHAR(xelt.character),
                                       CHAR(yelt.character)) >= 0;
            setElt(ansStorage, ansmethods, i, &xelt, &warn);
        }
        break;
    }
    CoercionWarning(xwarn);
    CoercionWarning(ywarn);
    CoercionWarning(warn);
}

static SEXP
externalVectorCompare(SEXP genericName, SEXP x, SEXP sxmethods,
                      SEXP y, SEXP symethods)
{
    SEXP xstorage = GET_SLOT(x, install("storage"));
    int nx = asInteger(GET_SLOT(xstorage, install("length")));
    SEXP ystorage = GET_SLOT(y, install("storage"));
    int ny = asInteger(GET_SLOT(ystorage, install("length")));
    storageMethods_t* xmethods =
        (storageMethods_t*) R_ExternalPtrAddr(sxmethods);
    storageMethods_t* ymethods =
        (storageMethods_t*) R_ExternalPtrAddr(symethods);
    int mode;
    int n, iop;
    SEXP ans, ansStorage;
    storageMethods_t* ansmethods;

    if (nx == 0 || ny == 0)
        return allocVector(LGLSXP, 0);

    genericName = STRING_ELT(genericName, 0);
    iop = genericName2opcode(CHAR(genericName), LENGTH(genericName));

    if (xmethods->mode == STRSXP || ymethods->mode == STRSXP)
        mode = STRSXP;
    else if (xmethods->mode == CPLXSXP || ymethods->mode == CPLXSXP)
        mode = CPLXSXP;
    else if (xmethods->mode == REALSXP || ymethods->mode == REALSXP)
        mode = REALSXP;
    else mode = INTSXP;
    n = (nx < ny)?ny:nx;
    if (n == nx)
        ans = PROTECT(xmethods->alloc(xstorage, allocVector(LGLSXP, 1), n));
    else ans = PROTECT(ymethods->alloc(ystorage, allocVector(LGLSXP, 1), n));
    ansStorage = GET_STORAGE_EXT_PTR(ans);
    xstorage = GET_STORAGE_EXT_PTR(xstorage);
    ystorage = GET_STORAGE_EXT_PTR(ystorage);
    ansmethods = getStorageMethods(ans);
    switch (mode) {
    case STRSXP:
        characterCompare(iop, nx, xstorage, xmethods,
                         ny, ystorage, ymethods,
                         n, ansStorage, ansmethods);
        break;
    case CPLXSXP:
        complexCompare(iop, nx, xstorage, xmethods,
                       ny, ystorage, ymethods,
                       n, ansStorage, ansmethods);
        break;
    case REALSXP:
        numericCompare(iop, nx, xstorage, xmethods,
                       ny, ystorage, ymethods,
                       n, ansStorage, ansmethods);
        break;
    case INTSXP:
        integerCompare(iop, nx, xstorage, xmethods,
                       ny, ystorage, ymethods,
                       n, ansStorage, ansmethods);
        break;
    }
    UNPROTECT(1);
    return ans;
}

static SEXP
externalVectorUnaryMinus(SEXP x, SEXP sxmethods,
                         SEXP ans, SEXP sansmethods)
{
    storageMethods_t* xmethods =
        (storageMethods_t*) R_ExternalPtrAddr(sxmethods);
    storageMethods_t* ansmethods =
        (storageMethods_t*) R_ExternalPtrAddr(sansmethods);
    getExtElt_t getElt;
    setExtElt_t setElt = ansmethods->setExtElt;
    int i;
    int n = asInteger(GET_SLOT(GET_SLOT(x, install("storage")),
                               install("length")));
    SEXP xstorage = GET_STORAGE_EXT_PTR(GET_SLOT(x, install("storage")));
    SEXP ansStorage = GET_STORAGE_EXT_PTR(GET_SLOT(ans, install("storage")));
    primitiveElt_t elt;
    int warn = 0;

    switch (ansmethods->mode) {
    case INTSXP:
        getElt = xmethods->integerGetElt;
        for (i = 0; i < n; i++) {
            getElt(xstorage, xmethods, i, &elt, &warn);
            if (elt.integer != NA_INTEGER)
                elt.integer = -elt.integer;
            setElt(ansStorage, ansmethods, i, &elt, &warn);
        }
        break;
    case REALSXP:
        getElt = xmethods->numericGetElt;
        for (i = 0; i < n; i++) {
            getElt(xstorage, xmethods, i, &elt, &warn);
            if (!ISNAN(elt.numeric))
                elt.numeric = -elt.numeric;
            setElt(ansStorage, ansmethods, i, &elt, &warn);
        }
        break;
    case CPLXSXP:
        getElt = xmethods->complexGetElt;
        for (i = 0; i < n; i++) {
            getElt(xstorage, xmethods, i, &elt, &warn);
            if (!ISNAN(elt.cmplx.r) && !ISNAN(elt.cmplx.i)) {
                elt.cmplx.r = -elt.cmplx.r;
                elt.cmplx.i = -elt.cmplx.i;
            } else if (!ISNAN(elt.cmplx.i)) {
                elt.cmplx.i = elt.cmplx.r;
            } else if (!ISNAN(elt.cmplx.r)) {
                elt.cmplx.r = elt.cmplx.i;
            }
            setElt(ansStorage, ansmethods, i, &elt, &warn);
        }
        break;
    }
    return ans;
}

static SEXP
externalVectorSummary(SEXP args)
{
    SEXP genericName;
    int nameLength;
    int narm;
    SEXP ansList;
    int iop;
    SEXP last_other, last_ext, last_obj, otherList, objList;

    if (length(args) < 3)
        error("insufficient number of arguments");
    args = CDR(args);
    PROTECT(args);
    genericName = asChar(CAR(args));
    args = CDR(args);
    nameLength = LENGTH(genericName);
    narm = asLogical(CAR(args));
    args = CDR(args);
    iop = genericName2opcode(CHAR(genericName), nameLength);
    UNPROTECT(1);
    if (iop == UNKNOWN_GENERIC)
        return args;
    PROTECT(args);
    last_other = otherList = CONS(R_NilValue, R_NilValue);
    PROTECT(last_other);
    last_obj = objList = CONS(R_NilValue, R_NilValue);
    PROTECT(last_obj);
    last_ext = ansList = CONS(R_NilValue, R_NilValue);
    PROTECT(last_ext);
    for (; args != R_NilValue; args = CDR(args)) {
        SEXP x = CAR(args);
        if (isObject(x)) {
            if (isExternalVectorWithStorage(x)) {
                SEXP xstorage = GET_SLOT(x, install("storage"));
                storageMethods_t* xmethods = getStorageMethods(xstorage);
                int n = asInteger(GET_SLOT(xstorage,
                                           install("length")));
                SEXP ans;

                if (n == 0) {
                    ans = allocVector(xmethods->mode, 0);
                } else {
                    xstorage = GET_STORAGE_EXT_PTR(xstorage);
                    if (iop <= MIN_GENERIC) {
                        ans = externalVectorMaxMin(iop, n, xstorage,
                                                   xmethods, narm);
                    } else if (iop == RANGE_GENERIC) {
                        ans = externalVectorRange(n, xstorage,
                                                  xmethods, narm);
                    } else if (iop <= SUM_GENERIC) {
                        ans = externalVectorSumProd(iop, n, xstorage,
                                                    xmethods, narm);
                    } else {
                        ans = externalVectorAllAny(iop, n, xstorage,
                                                   xmethods, narm);
                    }
                    args = CONS(ans, R_NilValue);
                    PROTECT(args);
                    SETCDR(last_ext, args);
                    last_ext = args;
                    UNPROTECT(1);
                }
            } else {
                args = CONS(x, R_NilValue);
                PROTECT(args);
                SETCDR(last_obj, args);
                last_obj = args;
                UNPROTECT(1);
            }
        } else {
            args = CONS(x, R_NilValue);
            PROTECT(args);
            SETCDR(last_other, args);
            last_other = args;
            UNPROTECT(1);
        }
    }
    SETCDR(last_ext, R_NilValue);
    if (last_obj != objList) {
        if (last_other != otherList) {
            SETCDR(last_obj, CDR(otherList));
            last_obj = last_other;
        }
        SETCDR(last_obj, CDR(ansList));
        ansList = CDR(objList);
    } else if (last_other != otherList) {
        SETCDR(last_other, CDR(ansList));
        ansList = CDR(otherList);
    } else {
        ansList = CDR(ansList);
    }
    UNPROTECT(4);
    return coerceVector(ansList, VECSXP);
}

#include <R.h>
#include <R_ext/Rdynload.h>

static const R_CallMethodDef R_CallDef  [] = {
    CALL_DEF(externalVectorMath, 5),
    CALL_DEF(externalVectorComplex, 5),
    CALL_DEF(externalVectorBinaryArith, 5),
    CALL_DEF(externalVectorCompare, 5),
    CALL_DEF(externalVectorUnaryMinus, 4),
    {NULL, NULL, 0},
};

static const R_ExternalMethodDef R_ExternalDef  [] = {
    EXTERNAL_DEF(externalVectorSummary),
    {NULL, NULL, 0},
};

const R_CallMethodDef*
groups_Call_methods(int* nptr)
{
    *nptr = LENGTH_OF_METHODS_ARRAY(R_CallDef);
    return &R_CallDef[0];
}

const R_ExternalMethodDef*
groups_External_methods(int* nptr)
{
    *nptr = LENGTH_OF_METHODS_ARRAY(R_ExternalDef);
    return &R_ExternalDef[0];
}
